
#include <types.h>
#include <lib.h>
#include <synch.h>
#include <thread.h>
#include <kern/errno.h>
#include "channel.h"

#define CHANNEL_SIZE 64
#define MAX_DATA_LEN 48 /* includes '\0' */

/* Global semaphore for cleaning up when finished */
static struct semaphore *finished_semaphore;

/* An example of a channel item type --- a string */
static void *string_item_create(void)
{
    return kmalloc(MAX_DATA_LEN);
}

static void string_item_copy(void *dest, void *src)
{
    strcpy(dest, src);
}

static void string_item_destroy(void *item)
{
    kfree(item);
}

static struct channel_item_ops string_ops = {
    string_item_create,
    string_item_copy,
    string_item_destroy
};


/* This function starts a number of threads */
static void start_threads(int nthreads, void *handle, void (*func)(void *, unsigned long))
{
    int i;
    for (i = 0; i < nthreads; i++) {
	int error = thread_fork("channel thread", handle, i, func, NULL);
	/*
	 * panic() on error.
	 */
	
	if (error) {
	    panic("start_threads: thread_fork failed: %s\n", strerror(error));
	}
    }
}

static void reader_thread(void *handle, unsigned long tnum)
{
    char buf[MAX_DATA_LEN];
    int ret;

    while((ret = channel_read(handle, buf)) == SUCCESS) 
	kprintf("reader %lu: got '%s')\n", tnum, buf);

    kprintf("reader %lu: Finished (%s)\n", tnum,
	    ret == FINISHED ? "OK" : "ERROR");

    V(finished_semaphore);
}


static void writer_thread(void *handle, unsigned long tnum)
{
    char buf[MAX_DATA_LEN];

    snprintf(buf, MAX_DATA_LEN, "Hello, I am writer %lu", tnum);
    channel_write(handle, buf);

    channel_finished(handle);

    V(finished_semaphore);
}

int cmd_channel (int nargs, char ** args)
{
    int nreaders, nwriters, i;
    void *handle;

    if(nargs == 1)
	nreaders = nwriters = 1;
    else if(nargs == 3) {
	nreaders = atoi(args[1]);
	nwriters = atoi(args[2]);
    } else {
	kprintf("Usage: chan [nreaders nwriters]\n");
	return EINVAL;
    }
    
    if(channel_create(nreaders, nwriters, CHANNEL_SIZE,
		      &string_ops, &handle) != SUCCESS) {
	kprintf("channel_create failed.\n");
	return EUNIMP;
    }

    finished_semaphore = sem_create("Finished", 0);
    if(!finished_semaphore) {
	kprintf("sem_create failed.\n");
	return EUNIMP; /* Probably not the best error to use, but is most suitable */
    }

    /* Start reader and writer threads */
    start_threads(nreaders, handle, reader_thread);
    start_threads(nwriters, handle, writer_thread);

    /* Wait for all threads to finish */
    for(i = 0; i < nreaders + nwriters; i++) 
	P(finished_semaphore);

    return 0;
}
	    

