

#include <types.h>
#include <lib.h>
#include <synch.h>
#include <curthread.h>
#include <thread.h>
#include <kern/errno.h>
#include <machine/spl.h>

#define MAX_DATA_LEN 255 /* includes '\0' */
#define MAX_THREAD_COUNT 2

/* function pointer for theads now have a type */
typedef void (*threadfp)(void * arg1, unsigned long arg2);

/* to idenfity when all threads have finished. */
struct semaphore * thread_finish_sem ;

/*
 * The 2 resources  
 */
struct simple_channel *Aschannel ;
struct simple_channel *Bschannel ;

/* Table of threads */
struct thread **thread_table ;



/* an extremely simple channel which can queue only one message */
struct simple_channel {
    const char *payload;		/* a string passed by reference */
};

/* send a message on a channel without blocking */
void
simple_channel_send(struct simple_channel *handle,
		            const char *payload)
{
    int spl; 

    /* assume for this example that we have exclusive access to the channel */
    handle->payload = payload;

	spl = splhigh();
	/* wake up anyone who is expecting a message from me */
    thread_wakeup(curthread);
    splx(spl);
}

/* receive a message from the the thread specified by "from" */
const char *
simple_channel_receive_from(struct thread *from,
      		                struct simple_channel *handle)
{
    int spl;
    
	spl = splhigh();
    thread_sleep(from); 	/* wait for "from" to send me a message */
    splx(spl);

    return handle->payload;
}


/*
 * Thread A 
 *
 * Thread A attempts to receive a message from thread B 
 * and then send a message to thread B.  
 * 
 * receiving is done in A's channel and sending is done on B's channel.
 */
void 
simple_thread_A(void * arg1, unsigned long arg2)
{

    struct thread * tB ; 
    const char *message;
    /* tell the compiler to go get lost */
    (void) arg1;
    (void) arg2;


    kprintf("hello I'm thread A\n");
    
    /* the other thread */
    tB = thread_table[1];
 

    /* recive from thread B */
    message = simple_channel_receive_from(tB,Aschannel);
    kprintf("%s:I got sent ('%s')\n",curthread->t_name, message);



    kprintf("%s:I'm sending to %s\n",curthread->t_name,tB->t_name);
    /* send to thread B */
    simple_channel_send(Bschannel, "Hello B, this is A!");

    /* sleep for a second */
    clocksleep(1);
    /* tell the main thread we're done */
    V(thread_finish_sem);
}

/*
 * Thread B 
 *
 * Thread B attempts to receive a message from thread A 
 * and then send a message to thread A.  
 * 
 * receiving is done in B's channel and sending is done on A's channel.
 */
void 
simple_thread_B(void * arg1, unsigned long arg2)
{

    
    
    struct thread * tA ; 
    const char *message;
    /* tell the compiler to go get lost */
    (void) arg1;
    (void) arg2;

    kprintf("hello I'm thread B\n"); 
    
    /* the other thread */
    tA = thread_table[0];

    
    /* receive from thread A */ 
    message = simple_channel_receive_from(tA,Bschannel);
    kprintf("%s:I got sent ('%s')\n",curthread->t_name,message);
  
    
    kprintf("%s:I'm sending to %s\n",curthread->t_name,tA->t_name);
    /* send to thread A */
    simple_channel_send(Aschannel, "Hello A, this is B!");


 

    /* sleep for a second */
    clocksleep(1);
    /* tell the main thread we're done */
    V(thread_finish_sem);
}






/*
 * main function that gets called from the menu.
 *
 */
int cmd_simple_sync (int nargs, char ** args)
{
    (void)nargs;
    (void)args;
    int error;
    int i;
    int nthreads = MAX_THREAD_COUNT ;
    
 
    threadfp thread_func_table[] = {
	/* theads */
	simple_thread_A,
	simple_thread_B
    };

    /* allocate memory for data structs */
    thread_table = kmalloc(sizeof(struct thread *) *MAX_THREAD_COUNT );
    Aschannel    = kmalloc(sizeof(struct simple_channel)  );
    Bschannel    = kmalloc(sizeof(struct simple_channel)  );
  
    /* say hello to the world */
    kprintf("simple sync program \n");
    
    
    /* 
     * create all semaphores we need 
     */ 
    thread_finish_sem = sem_create("Finished", 0);
    if(NULL ==  thread_finish_sem) {
	kprintf("sem_create failed.\n") ;
        /* Probably not the best error to use, but is most suitable */
	return EUNIMP; 
    }



    /*
     * start up all the threads 
     */
    
    /* start up thread A*/
    error = thread_fork("thread A",NULL,-1,thread_func_table[0],
			&(thread_table[0]) );
    kprintf("started %s\n",thread_table[0]->t_name); 
    /* panic() on error. */
    if (error)
	panic("start_threads: thread_fork failed: %s\n",
	      strerror(error));
    
    
    
    /* start up thread B*/
    error = thread_fork("thread B",NULL,-1,thread_func_table[1],
			&(thread_table[1]) );
    kprintf("started %s\n",thread_table[1]->t_name);
    
    /* panic() on error. */
    if (error)
	panic("start_threads: thread_fork failed: %s\n",
	      strerror(error));
    
    


    
    /* Wait for all threads to finish */
    for(i = 0; i < nthreads; i++) 
	P(thread_finish_sem);
    
    
    return 0;
}


